//=============================================================================


#pragma once


//== INCLUDES =================================================================


#include "Energy_term.h"

#include "../my_types.h"


//== NAMESPACE ================================================================


namespace graphene {
namespace surface_mesh {


//== CLASS DEFINITION =========================================================


class
Energy_term_pca_landmarks_skinned_multiple_scans : public Energy_term
{
public:

    Energy_term_pca_landmarks_skinned_multiple_scans(const Energy_term_data& energy_term_data, double weight);

    virtual ~Energy_term_pca_landmarks_skinned_multiple_scans();

    virtual void pre_minimize_action();

    virtual double evaluate();

    virtual void add_rows_big_lse( std::vector<Tripl>& coeffs,
                                   Eigen::VectorXd& b,
                                   unsigned int& r );

    virtual unsigned int n_rows_big_sle();

    virtual bool small_sle_supported() { return false; }

    virtual bool requires_two_pointsets_and_posevectors() {return true;}

    virtual bool needs_pca() {return true;}


    virtual void solve_result(std::set<Solve_result>& solve_results) {solve_results.insert(SR_PCA);}


private:

    struct Per_fitting_target_data
    {
        std::vector<Point> lm_ps;
        std::vector<Surface_mesh::Vertex> lm_templ_vertices;

        Surface_mesh::Vertex_property<double> local_landmark_weight;
    };
};


//=============================================================================
} // namespace surface_mesh
} // namespace graphene
//=============================================================================
